#
# This is a simple GUI for browsing S4-style classes.
# Specifically, it allows one to look at all the
# classes, their (single) inheritance, slots and 
# methods that are defined for them.
#
# Add a menu item for controlling how the slots are displayed (combined == TRUE | FALSE).
# 

showClassSlots <-
function(className, html)
{
  # Really want a way to do toHTML(getSlots())
  # and perhaps put in additional information.

   # Get the contents of the HTML to display.
  txt <- getSlotHTMLTable(className, TRUE, TRUE)

   # Now load it.
  txt <- paste("<html><body>", txt, "</body></html>")  
  html$LoadFromString(txt, -1)

  invisible(txt)
}

getSlotHTMLTable <-
function(className, complete = TRUE, combined = FALSE)
{
  #
  # Compute the HTML text for displaying the slot information for 
  # the specified class.
  # The complete and combined arguments control if and how we group
  # information about inherited slots.
  # If combined is FALSE, then we get separate tables for each
  # of the ancestor/base classes in the inheritance hierarchy

  # Alternatively, if combined is TRUE, then if complete is TRUE,
  # we put all the slots - inherited and defined locally with the class -
  # together into a single table, but color code each slot by the class
  # in which it was defined.
  # See formatClassSlots.
  
 if(combined == FALSE) {
   paste(sapply(c(className, getAllSuperClasses(getClass(className))),
                function(x) formatClassSlots(getSlots(x, complete = FALSE), x)), collapse="<br>")
 } else {
   if(complete) {
     classNames <- c(className, getAllSuperClasses(getClass(className)))
     slots <- lapply(classNames, function(x) getSlots(x, complete = FALSE))
     names(slots) <- classNames
     formatClassSlots(slots, className)
   } else  {
     slots <- getSlots(className, complete = FALSE)
     formatClassSlots(slots, className)
   }
 }
}  

formatClassSlots <-
function(slots, className)
{
 if(length(slots) == 0)
   return(paste("No Slots in `", className, "' class\n", sep=""))

 header <- "<tr><td>Name</td><td>Type</td></tr>" 

 doSlot <-
   function(name, slots, className = "", color = "black") {
     paste("<tr align='left'><th>",
            "<font color='", color, "'>", name, "</font>", "</th><th>",
           "<a href='", slots[name], "'>", slots[name],"</a>",
           "</th></tr>", collapse="", sep="")
     }

  # Want to color code the slots based on where they were inherited from.
  # or create a table with the slots from the parent classes in a different
  # part of the table. 
 if(is.list(slots)) {
   colors <- c("red", "blue", "brown", "green")
   contents <- sapply(1:length(slots),
                      function(x) {
                        if(length(slots[[x]]))
                           # use indices rather than names so that we can index the colors!
                          paste(sapply(1:length(slots[[x]]),
                                       function(i) {
                                         z <- names(slots[[x]])[i]
                                         doSlot(z, slots[[x]], names(slots)[x], color = colors[x%%length(colors)])
                                     }), collapse="\n")
                        else ""})

   num <- length(slots)
   hierarchy <- paste("<pre>",
                       paste(sapply(num:1,
                                    function(i)
                                     paste(
                                           paste(rep(" ", num - i), collapse=""),
                                           "<a href='", names(slots)[i], "'>",
                                           "<font color='",
                                           colors[i%%length(colors)],
                                           "'>",
                                           names(slots)[i],
                                           "</font>", "</a>", collapse="",sep="")),
                             collapse="\n"),
                      "</pre>", sep="\n", collapse="")

 } else  {
   contents <- sapply(names(slots), function(x) doSlot(x, slots))
   hierarchy <- ""
 }

 paste("<H2>", className, "</H2>",
       hierarchy,
       "<table border='1'>",
       header,
       paste(contents, collapse="\n"),
       "</table>",
       sep="\n",
       collapse="\n")
}  

S4ClassView <-
function(classes = getClasses(where = where), where = .GlobalEnv)
{
  require(RGtk)
  require(RGtkHTML)
  
  win <- gtkWindow(show = FALSE)
  win$SetTitle("S4 Class & Method Viewer")
  
  h <- gtkHPaned()

  hier <- getS4ClassHierarchy(classes)

  callback <- function(className, w) {
      # Ignore the artificially created top-level class.
    if(className == "SObject")
      return()    
    showClassSlots(className, html)
    methods <- getClassMethods(className)
    if(length(methods))
      showClassMethods(methods, glist, showMethodHTML, className)
  }

  showMethodHTML <-
    function(d, w) {
      methodName <- d[[1]]
      className <- d[[2]]
      mlist <- d[[3]]

      methods <- mlist[[methodName]] 
      txt <- paste("<h3>Methods for", methodName, "involving", className,"</h3>",
                   "Number of methods: ", length(methods))
      methodsHtml$LoadFromString(txt, -1)    
    }
  
  tr <- createInheritanceTree(hier, callback = callback)

  sw <- gtkScrolledWindow()
  sw$AddWithViewport(tr$tree)
  sw$SetUsize(200, 500)
  h$Add(sw)

  sw <- gtkScrolledWindow()
  html <- gtkHTML()

  html$AddCallback("link-clicked", function(w, link) {
                     f(link)
                   })
  
  sw$Add(html)

  book <- gtkNotebook()
  book$AppendPage(sw, gtkLabel("Classes & Slots"))

  sw <- gtkScrolledWindow()
  glist <- gtkList()
  sw$AddWithViewport(glist)
  sw$SetUsize(90, 600)  

  mh <- gtkHPaned()
  mh$Add(sw)

  sw <- gtkScrolledWindow()
  methodsHtml <- gtkHTML()
  sw$Add(methodsHtml)
  mh$Add(sw)
  
  book$AppendPage(mh, gtkLabel("Methods"))

  h$Add(book)

  win$Add(h)
  win$SetUsize(500, 600)  
  win$Show()

  invisible(win)
}

getS4ClassHierarchy <-
function(classes = getClasses())
{
  descendants <- list()

  sapply(classes,
          function(name) {
            which <- names(getExtends(getClass(name))[1])
            if(is.null(which))
              which <- "NA"
            if(is.na(match(which, names(descendants))))
               descendants[[which]] <<-  name
            else
               descendants[[which]] <<-  c(descendants[[which]], name)

            TRUE
          })


    # The particulars of the S4 class hierarchy is that there is no
    # single top-level class. Instead, we have many that have no ancestor.
    # So we just artificially create this.
   descendants[["SObject"]] <- descendants[["NA"]]
   descendants[["NA"]] <- "SObject"

   descendants
}  

getClassMethods <-
function(className, inherited = TRUE)
{
  generics <- getGenerics()
  classes <- c(className, getAllSuperClasses(getClass(className)))
  els <- lapply(generics,
                function(funName) {
                  mlist <- getMethods(funName)
                  mlist <- linearizeMlist(mlist, inherited = inherited)

                  if(length(mlist@classes))
                    mlist@methods[!sapply(mlist@classes, function(x, y) all(is.na(match(x, y))), classes)]
                  else
                    NULL
         })

  names(els) <- generics

  els[sapply(els, function(x) length(x) > 0)]
}


showClassMethods <-
function(mlist, glist, callback, className)
{
  glist$ClearItems(0, 100)
  lapply(names(mlist),
         function(x) {
            item <- gtkListItem(x)
            glist$Add(item)
            item$AddCallback("select", callback, list(x, className, mlist))
            item
          })

}  


