collapseClassName <-
  #
  # This is used to convert a name of the form GtkClass
  # to gtk_class and then it can be used to generate the
  # name of a Gtk routine for the object type.
function(name)
{  
  tmp <- gsub("([ABCDEFGHIJKLMNOPQRSTUVWXYZ]+)", "_\\1", name)
  tmp <- tolower(substring(tmp, 2))
  gsub("_([abcdefghijklmnopqrstuvwxyz])_","_\\1", tmp)  
}

getGtkInheritance <-
function(klasses = .GtkClasses, descendants = TRUE)
{
     ancestors <- list()

     for(i in klasses) {

       l <- paste(collapseClassName(i), "get_type", sep="_")
       if(!is.loaded(l))
        warning(l, "is not available")
       else {
          # call the gtk_<type>_get_type() routine to initialize the Gtk class.
         .C(l)
          # now get its ancestors (and itself)
         tmp <- gtkTypeGetClasses(gtkGetType(i))
          # Add it to the list.
         ancestors[[i]] <- tmp
       }
     }


     collect <-
       function(klasses) {
         h <- list()
         sapply(names(klasses),  function(x) {
           tmp <- klasses[[x]]
           if(length(klasses) > 1) {
             h[[tmp[2]]] <<- c(h[[tmp[2]]], x)
           }
           TRUE
         })
         h
       }

     if(descendants)
        val <- collect(ancestors)
     else
        val <- ancestors

     val
}

createInheritanceTree <-
function(hier, el = hier[["NA"]][1], top= gtkTree(), callback = NULL)
{
  l <- gtkTreeItem(el)
  if(!is.null(callback))
    l$AddCallback("select", callback, el)
  top$Append(l)

  if(length(hier[[el]]) == 0) {
    return(list(item = l, subTree = NULL, tree = top))
  }
  
  tr <- gtkTree(show = FALSE)
  for(i in hier[[el]]) {
    tmp <- createInheritanceTree(hier, i, tr, callback)
  }
  l$SetSubtree(tr)

  list(item = l,  subTree = tr, tree = top)
}

showGtkInheritance <-
function(klasses = .GtkClasses)
{
  properties <- gtkHTMLNew()
  class(properties) <- gtkObjectGetClasses(properties, FALSE)

  signals <- gtkHTMLNew()
  class(signals) <- gtkObjectGetClasses(signals, FALSE)
  
  showInfo <-
    function(cmd, w = NULL) {
       info <- getPropertyInfo(cmd, strip = TRUE)
       txt <- propertyInfoToHTML(info, cmd)
       properties$LoadFromString(txt, -1)

       info <- gtkTypeGetSignals(cmd)
       txt <- signalInfoToHTML(info, cmd)
       signals$LoadFromString(txt, -1)
       TRUE
    }

  handleLink <-
    function(w, link) {
      if(length(grep("^R:", link))) {
        eval(parse(text = substring(link, 3)))
      } else
      showInfo(link, NULL)
    }
  signals$AddCallback("link-clicked", handleLink)
  properties$AddCallback("link-clicked", handleLink)    
  
  h <- getGtkInheritance(klasses)
  tr <- createInheritanceTree(h, callback = showInfo)

  win <- gtkWindow(show = FALSE)

  p <- gtkHPaned()
  
  sw <- gtkScrolledWindow()
  sw$AddWithViewport(tr$tree)
  p$Add(sw)


  book <- gtkNotebook()
  sw <- gtkScrolledWindow(show = FALSE)
  sw$Add(properties)
  book$AppendPage(sw, gtkLabel("Properties"))

  sw <- gtkScrolledWindow(show = FALSE)
  sw$Add(signals)
  book$AppendPage(sw, gtkLabel("Signals"))  
  
  p$Add2(book)
  p$SetPosition(250)
  win$Add(p)
  win$SetUsize(500, 200)
  win$ShowAll()

  win
}  

getPropertyInfo <-
function(type, ...)
{
  t <- gsub("Gtk", "gtk", type)
  n <- paste(t, "New", sep="")
  if(exists(n)) {
    val <- try(get(n)())
    if(inherits(val, "try-error"))
      return(NULL)
  }

  info <- gtkObjectGetArgInfo(type, ...)

  info
}  

propertyInfoToHTML <-
function(info, type)
{
 txt <- sapply(names(info),
               function(i) {
                 if(length(info[[i]]) == 0)
                   return("")

                 paste("<dl>", "<h3><a href=\"R:showInfo('", i, "')\">", i, "</a></h3",
                 paste(sapply(names(info[[i]]),
                              function(x, els) {
                                paste("<dt>", x ,"<dd>",
                                      "<a href=R:showInfo('",names(els[[x]]$type), "')>", names(els[[x]]$type), "</a>",
                                      "<br>",
                                      paste(names(.GtkArgFlags)[bitAnd(els[[x]]$flag, .GtkArgFlags)], collapse=", "), collapse="\n", sep="")
                              }, els = info[[i]]), collapse="\n"),
                    "</dl>", collapse="\n", sep="")
              })

 txt <- paste(txt, collapse="\n")


 k <- gtkTypeGetClasses(gtkGetType(type))
 prefix <- sapply(seq(2, length = length(k), by = 4), function(x) paste(rep(" ", x), collapse=""))

 hier <- c("<pre>", paste(prefix, sapply(rev(k),
                                         function(x) {
                                           paste("<a href=\"R:showInfo('",x,"')\">", x, "</a>", " (",length(info[[x]]), ")", sep="")
                                         }), sep="", collapse="\n")  , "</pre>")
 hier <- paste(hier, collapse="\n")


 paste("<HTML><BODY>\n", "<h2><a href='", type,"'>", type ,"</a></h2>", hier, txt, "</BODY></HTML>", collapse="\n", sep="")
}

signalInfoToHTML <-
function(signals, type)
{
 val <-
   sapply(names(signals),
        function(x) {
            info <- gtkSignalGetInfo(signals[[x]])
#print(info$objectType)
            tmp <- vector("list", length(info$parameters)+1)
            tmp[[1]] <- info$objectType
            if(length(info$parameters))
              tmp[2:length(tmp)] <- info$parameters
            parms <- sapply(tmp, #info$parameters,
                            function(x) {
                                 c("<th>", "<a href=R:showInfo('", names(x),"')>", names(x),"</a>", "</th>", sep="")
                            })

            parms <- c("Number of arguments: ", length(tmp), "<br>\n", "<table>",
                       "<th>(</th>", parms, "<th>)</th>",
                       "</table>", collapse="")

          txt <- c("<h4>", x, "</h4>\n",
                   "<a href=R:showInfo('", names(info$objectType),"')>", names(info$objectType), "</a>\n",
                   "<br>",
                   "\n",
                   parms,
                   "\n<br>\n",
                   "return value type: ", "<a href=R:showInfo('", names(info$returnType),"')>", names(info$returnType), "</a>",
                   "<br>\n",
                  )
          txt <- paste(txt, collapse="")
        })

  paste("<HTML><BODY>\n", paste(val, collapse="\n"), "\n</BODY></HTML>\n", collapse="\n")
}  
